import com.atlassian.jira.component.ComponentAccessor
import com.atlassian.jira.issue.search.SearchRequest
import com.atlassian.jira.issue.search.SearchRequestManager
import com.atlassian.jira.component.ComponentAccessor
import com.atlassian.query.clause.*
import org.apache.poi.ss.util.CellReference
import com.google.common.collect.LinkedListMultimap


// check dependency
if (!binding.variables.containsKey("fieldHelper")) {
	throw new Exception("The script \"issue-navigator.groovy\" requires the prior execution of \"field-helper-tool.groovy\".\nUse this expression in one of the topmost cells: <mt:execute script=\"field-helper-tool.groovy\"/><mt:execute script=\"issue-navigator.groovy\"/>")
}

// set these explicitly to avoid "not a property" exceptions
if (!binding.variables.containsKey("xlsView")) {
	xlsView = null
}
if (!binding.variables.containsKey("automationTriggerJiraJqlExpression")) {
	automationTriggerJiraJqlExpression = null
}
if (!binding.variables.containsKey("searchRequest")) {
	searchRequest = null
}
if (!binding.variables.containsKey("tempo")) {
	tempo = null
}

if(searchRequest?.name != null) {
	filename = "${searchRequest.name}.xlsx".toString() // "filename" is a standard Java String, not a GString
}

issueNavigator = new IssueNavigatorTool(xlsView: xlsView, user: user, searchRequest: searchRequest,
		tableLayoutFactory: tableLayoutFactory,	fieldHelper: fieldHelper, tempo: tempo,
		automationTriggerJiraJqlExpression: automationTriggerJiraJqlExpression
)


/**
 * Tool to export dynamic column configurations from Issue Navigator.
 * It requires the prior execution of "field-helper-tool.groovy".
 */
public class IssueNavigatorTool {
	def DEFAULT_COLUMNS = [ "issuetype", "issuekey", "summary", "assignee", "reporter", "priority", "status", "resolution", "created", "updated", "duedate" ]

	def searchRequestManager = ComponentAccessor.getOSGiComponentInstanceOfType(SearchRequestManager.class)
	def tableLayoutFactory // cannot be obtained from ComponentAccessor
	def worklogManager = ComponentAccessor.worklogManager

	def xlsView
	def user
	def searchRequest

	def fieldHelper
	def issueTableLayout
	def tempo

	def automationTriggerJiraJqlExpression

	/**
	 * Returns whether the all columns should be exported (or the selected ones only).
	 * @see http://www.midori-global.com/products/jira-better-excel-plugin/documentation/recipes#all-fields-vs-selected-fields
	 */
	private isAllFieldsMode() {
		return viewNameContains("all")
	}

	private viewNameContains(s) {
		return (boolean) xlsView?.name?.toLowerCase()?.contains(s)
	}

	def getIssueTableLayout() {
		// lazily calculate the table layout
		if(issueTableLayout == null) {
			// if automation is used, find the search request (saved filter) object based on the automation trigger's JQL
			def jql = automationTriggerJiraJqlExpression
			if (jql != null) {
				def filterIdOrName = parseFilterIdOrNameFromJql(jql)
				if (filterIdOrName != null) {
					if (filterIdOrName.isInteger()) {
						// find by ID
						searchRequest = searchRequestManager.getSearchRequestById(Long.parseLong(filterIdOrName))
					}
					if (searchRequest == null) {
						// find by name
						def searchRequestsFound = searchRequestManager.findByNameIgnoreCase(filterIdOrName)
						if (!searchRequestsFound.empty) {
							searchRequest = searchRequestsFound.get(0)
						}
					}
				}
			}

			// get the all fields layout, the filter's column layout or a sensible default column layout
			def allFields = isAllFieldsMode()

			if(searchRequest != null) {
				issueTableLayout = allFields ? tableLayoutFactory.getAllColumnsExcelLayout(searchRequest, user) : tableLayoutFactory.getStandardExcelLayout(searchRequest, user)
			} else {
				issueTableLayout = allFields ? tableLayoutFactory.getAllColumnsExcelLayout(new SearchRequest(), user) : tableLayoutFactory.getDashboardLayout(user, DEFAULT_COLUMNS)
			}
		}

		return issueTableLayout
	}

	def getFieldValue(issue, columnLayoutItem) {
		fieldHelper.getFieldValue(issue, columnLayoutItem.navigableField, columnLayoutItem)
	}

	def getFieldCellType(columnLayoutItem) {
		fieldHelper.getFieldCellType(columnLayoutItem.navigableField)
	}

	def getFieldFormat(columnLayoutItem) {
		fieldHelper.getFieldFormat(columnLayoutItem.navigableField)
	}

	/**
	 * Returns the "aggregation" formula to be inserted into the last row in the given column, or null if the column should not be aggregated.
	 * This is typically a formula using SUM(), COUNTA() or such.
	 */
	def getFieldSumFormula(columnLayoutItem, cell) {
		def cellReference = new CellReference(cell)
		def firstDataRowCellReferenceString = new CellReference(1, cell.columnIndex).formatAsString()
		def lastDataRowCellReferenceString = new CellReference(cell.rowIndex - 1, cell.columnIndex).formatAsString()

		!fieldHelper.isCustomField(columnLayoutItem.navigableField) ? getSystemFieldSumFormula(columnLayoutItem.navigableField, firstDataRowCellReferenceString, lastDataRowCellReferenceString) : getCustomFieldSumFormula(columnLayoutItem.navigableField, firstDataRowCellReferenceString, lastDataRowCellReferenceString)
	}

	private getSystemFieldSumFormula(field, firstDataRowCellReferenceString, lastDataRowCellReferenceString) {
		switch(field.id) {
			case "issuekey":
				// count of issues
				"COUNTA(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			case "timeoriginalestimate":
			case "timeestimate":
			case "timespent":
			case "aggregatetimespent":
			case "aggregatetimeestimate":
			case "aggregatetimeoriginalestimate":
			case "votes":
			case "watches":
				// sum of values
				"SUM(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			default:
				// no formula
				null
		}

	}

	private getCustomFieldSumFormula(customField, firstDataRowCellReferenceString, lastDataRowCellReferenceString) {
		switch(customField.customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
				// sum of values
				"SUM(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			// JIRA Service Desk custom field types
			case "com.atlassian.servicedesk:sd-sla-field":
				// average of values
				"AVERAGE(${firstDataRowCellReferenceString}:${lastDataRowCellReferenceString})"
				break

			default:
				// no formula
				null
		}
	}

	/**
	 * Parses a saved filter's name or ID from a JQL query string (in the format: 'filter = 12' or 'filter = "My open bugs"').
	 * @return filter id or name without quotation marks or apostrophes, or null if wasn't found.
	 */
	private parseFilterIdOrNameFromJql(String jql) {
		// parse JQL
		def clazz = ComponentAccessor.class.classLoader.loadClass("com.atlassian.jira.jql.parser.JqlQueryParser")
		def jqlQueryParser = ComponentAccessor.getOSGiComponentInstanceOfType(clazz)
		def query = jqlQueryParser.parseQuery(jql);

		// search for the "filter =" clause
		def filterIdOrName = query.whereClause.accept(new ClauseVisitor<String>() {
			String visit(AndClause andClause) {
				visitSubClauses(andClause.clauses)
			}

			String visit(NotClause notClause) {
				visitSubClauses(notClause.clauses)
			}

			String visit(OrClause orClause) {
				visitSubClauses(orClause.clauses)
			}

			String visit(TerminalClause clause) {
				return (("filter".equals(clause.name)) && ("=".equals(clause.operator.displayString))) ? ((clause.operand.longValue != null) ? clause.operand.longValue : clause.operand.stringValue) : null
			}

			String visit(WasClause clause) {
				return null
			}

			String visit(ChangedClause clause) {
				return null
			}

			private String visitSubClauses(def clauses) {
				for(clause in clauses) {
					def filterIdOrName = visit(clause)
					if(filterIdOrName != null) {
						return filterIdOrName;
					}
				}
				return null
			}
		})

		return filterIdOrName
	}

	def getWorklogs(issue) {
		def worklogs = worklogManager.getByIssue(issue)
		return worklogs.empty ? [[:]] : worklogs
	}

	def getTempoWorklogs(issue) {
		def worklogs = tempo.getWorklogs(issue.key)
		return worklogs.empty ? [[:]] : worklogs
	}

	/**
	 * Returns a "repeating" list of entries where the issue is the (repeating) key and its worklog is the value.
	 * Example: (i1 -> w11, i1 -> w12, i1 -> w13, i2 -> w21, i3 -> w31, ...)
	 */
	def getRepeatingWorklogs(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each {
			it -> result.putAll(it, getWorklogs(it))
		}

		return result.entries()
	}

	/**
	 * Returns a "repeating" list of entries where the issue is the (repeating) key and its Tempo worklog is the value.
	 * Example: (i1 -> w11, i1 -> w12, i1 -> w13, i2 -> w21, i3 -> w31, ...)
	 */
	def getRepeatingTempoWorklogs(issues) {
		def result = LinkedListMultimap.create() // keeps insertion order for both keys and values

		issues.each {
			it -> result.putAll(it, getTempoWorklogs(it))
		}

		return result.entries()
	}
}
